import numpy as np

from stylo.shape.shape import Shape


class Ellipse(Shape):
    """An ellipse.

    Mathematically we can define an ellipse to be the set
    of points :math:`(x, y)` which satisfy:

    .. math::

        \\frac{(x - x_0)^2}{a^2} + \\frac{(y - y_0)^2}{b^2} \\leq r^2

    where:

    - :math:`(x_0, y_0)` is the center of the ellipse
    - :math:`a` is known as the semi major axis, larger values make the
      ellipse more elongated in the :math:`x`-direction
    - :math:`b` is known as the semi minor axis, larger values make the
      ellipse more elongated in the :math:`y`-direction
    - :math:`r` is the "radius" of the ellipse and controls the overall
      size of the ellipse
    """

    def __init__(self, x, y, a, b, r):
        self.x = x
        self.y = y
        self.a = a
        self.b = b
        self.r = r

    def __repr__(self):
        return "Ellipse(x={0.x}, y={0.y}, a={0.a}, b={0.b}, r={0.r})".format(self)

    def draw(self):

        r = self.r * self.r
        a = self.a
        b = self.b
        x0 = self.x
        y0 = self.y

        def ellipse(x, y):
            xc = x - x0
            yc = y - y0

            return (xc * xc) / a + (yc * yc) / b <= r

        return ellipse


class Circle(Ellipse):
    """
    Mathematically a circle can be defined as the set of all
    points :math:`(x, y)` that satisfy

    .. math::

    (x - x_0)^2 + (y - y_0)^2 = r^2

    This function returns another function which when given
    a point :code:`(x, y)` will return :code:`True` if that
    point is in the circle
    """

    def __init__(self, x, y, r):
        super().__init__(x, y, 1, 1, r)

    def __repr__(self):
        return "Circle(x={0.x},y={0.y},r={0.r})".format(self)


class Rectangle(Shape):
    """
    It's quite simple to define a rectangle, simply pick a
    point :math:`(x_0,y_0)` that you want to be the center
    and then two numbers which will represent the width and
    height of the rectangle.
    """

    def __init__(self, x, y, width, height):
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def __repr__(self):
        arg_string = "x={0.x},y={0.y},width={0.width},height={0.height}".format(self)
        return "Rectangle({})".format(arg_string)

    def draw(self):

        left = self.x - (self.width / 2)
        right = self.x + (self.width / 2)
        top = self.y + (self.height / 2)
        bottom = self.y - (self.height / 2)

        def rectangle(x, y):
            xs = np.logical_and(left < x, x < right)
            ys = np.logical_and(bottom < y, y < top)

            return np.logical_and(xs, ys)

        return rectangle


class Square(Rectangle):
    def __init__(self, x, y, size):
        super().__init__(x, y, size, size)

    def __repr__(self):
        return "Square(x={0.x},y={0.y},size={0.width})".format(self)


class Triangle(Shape):
    """
    A Triangle can be defined by picking three non-collinear points
    :math:`(a, b, c)` in the form of a tuple each.
    Points inside the triangle are determined using the
    barycentric coordinate system method, which is further explained
    `in this answer <https://math.stackexchange.com/a/1884485/294670>`_
    and the first method
    `here <http://totologic.blogspot.com/2014/01/accurate-point-in-triangle-test.html>`_.
    """  # noqa E501

    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c

    def __repr__(self):
        return "Triangle(a={0.a}, b={0.b}, c={0.c})".format(self)

    def get_q(self):
        """
        Returns the denominator for calculating both the barycentric
        coordinates, which absolute value is the double of the area
        of the triangle.
        """
        a = self.a
        b = self.b
        c = self.c
        q = -b[1] * c[0] + a[1] * (-b[0] + c[0]) + a[0] * (b[1] - c[1]) + b[0] * c[1]

        return q

    def get_s(self, x, y):
        """
        Returns one of the barycentric coordinates of the triangle using both
        :math:`a` and :math:`c` points.
        """
        a = self.a
        c = self.c
        sign = -1 if self.get_q() < 0 else 1
        s = (a[1] * c[0] - a[0] * c[1] + (c[1] - a[1]) * x + (a[0] - c[0]) * y) * sign
        return s

    def get_t(self, x, y):
        """
        Returns one of the barycentric coordinates of the triangle using both
        :math:`a` and :math:`b` points.
        """
        a = self.a
        b = self.b
        sign = -1 if self.get_q() < 0 else 1
        t = (a[0] * b[1] - a[1] * b[0] + (a[1] - b[1]) * x + (b[0] - a[0]) * y) * sign
        return t

    def draw(self):
        def triangle(x, y):
            sign = -1 if self.get_q() < 0 else 1
            first_condition = self.get_s(x, y) > 0
            second_condition = self.get_t(x, y) > 0
            third_condition = self.get_s(x, y) + self.get_t(x, y) < self.get_q() * sign

            return np.logical_and(
                first_condition, np.logical_and(second_condition, third_condition)
            )

        return triangle



from abc import ABC, abstractmethod

import PIL as P
import matplotlib.pyplot as plt


class Drawable:
    def __init__(self, domain, shape, color):
        self.domain = domain
        self.shape = shape
        self.color = color

    def __iter__(self):
        return iter([self.domain, self.shape, self.color])


class Image(ABC):
    def __call__(self, width, height, filename=None, plot_size=None):

        image = self._render(width, height)
        self.data = image

        if filename is not None:
            self._save(image, filename)
            return

        if plot_size is None:
            plot_size = 12

        fig, ax = plt.subplots(1, figsize=(plot_size, plot_size))
        ax.imshow(image)

        return fig

    def _save(self, image, filename):

        height, width, _ = image.shape

        pil_image = P.Image.frombuffer(
            "RGB", (width, height), image, "raw", "RGB", 0, 1
        )

        with open(filename, "wb") as f:
            pil_image.save(f)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @abstractmethod
    def _render(self, width, height):
        pass


def render_drawable(drawable, image_data):
    """Given a drawable, render it onto the given image data."""

    domain, shape, color = drawable
    domain = shape._apply_transform(domain)

    height, width, depth = image_data.shape

    parameters = shape.parameters
    values = domain[parameters](width, height)

    coords = {k: v for k, v in zip(parameters, values)}
    mask = shape(**coords)

    return color(mask, image_data=image_data)



"""This script automatically writes the example sections in the documentation."""
import os

import argparse
import pathlib
from importlib import import_module
import sys

EXAMPLE_TEMPLATE = """
.. {reference}:

{title}

.. image:: {imgpath}
   :width: 75%
   :align: center

{description}

.. literalinclude:: ../../..{srcpath}
   :language: python
   :start-after: # <example>
   :end-before: # </example>
   :dedent: 4
"""

INDEX_TEMPLATE = """
.. {reference}:

{title}

.. toctree::

   {files}

"""


def format_title(title):
    underline = "=" * len(title)
    return "{}\n{}".format(title, underline)


def write_index(index):

    for type, items in index.items():

        if len(items) == 0:
            continue

        files = "\n   ".join(items)
        index_file = "docs/{}/examples/index.rst".format(type)

        context = {
            "files": files,
            "title": format_title("Examples"),
            "reference": "_{}_examples".format(type),
        }
        rst = INDEX_TEMPLATE.format(**context)

        with open(index_file, "w") as f:
            f.write(rst)

        # print("Wrote {} index.".format(type))


def write_example_page(root, example, index):

    info = example.example_info
    srcpath = example.__file__.replace(root, "")
    reference = "_" + info["type"] + "_example_" + info["name"]

    # print("\t{}".format(info["name"]))

    # Update the index
    index[info["type"]].append(info["name"])

    context = {
        "title": format_title(info["name"].capitalize()),
        "imgpath": "/_static/examples/" + info["name"] + ".png",
        "description": example.__doc__,
        "srcpath": srcpath,
        "reference": reference,
    }

    rst = EXAMPLE_TEMPLATE.format(**context)

    dirname = pathlib.Path("docs/" + info["type"] + "/examples/")

    if not dirname.exists():
        os.mkdir(str(dirname))

    filename = os.path.join(str(dirname), info["name"] + ".rst")
    with open(filename, "w") as f:
        f.write(rst)


def find_examples(path):
    """Find the examples and import them."""

    dir = pathlib.Path(path)
    examples = []

    for example_path in dir.glob("*.py"):

        if example_path.name == "__init__.py":
            continue

        module_name = str(example_path).replace("/", ".").replace(".py", "")
        examples.append(import_module(module_name))

    return examples


def parse_arguments():

    parser = argparse.ArgumentParser(prog="exampledoc.py")
    parser.add_argument(
        "-e, --examples",
        nargs=1,
        type=str,
        dest="examples",
        required=True,
        help="the folder containing the example tests.",
    )

    return parser.parse_args()


def main():

    root = os.path.abspath(".")

    # Fix the path so that the test files can be imported.
    sys.path.insert(0, root)

    args = parse_arguments()
    index = {"using": [], "extending": [], "contributing": []}

    examples = find_examples(args.examples[0])

    print("Found {} examples".format(len(examples)))

    for example in examples:
        write_example_page(root, example, index)

    write_index(index)


if __name__ == "__main__":
    main()



import pytest


def define_benchmarked_example(name, example):

    import matplotlib

    matplotlib.use("Agg")

    image = example()

    @pytest.mark.parametrize("n", [512, 1024, 2048])
    def benchmark_test(benchmark, n):

        filename = None

        if n == 512:
            filename = "docs/_static/examples/" + name.lower() + ".png"

        benchmark(image, n, n, filename=filename)

    return benchmark_test
